"Python3 document categorization library, mawtoks method."

##  DOCALIB :
##
##  Python3 data & routines to supply to toccata test harness s/w :
##  Copyright 2013, R.S. Forsyth.
##  Released under GNU Public licence, version 3:
##  http://www.gnu.org/licenses/gpl.txt
##  first version : 13/08/2013
##  last revision : 27/03/2017


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)

import  math, os, string, sys

if sys.platform.startswith("linu"):  ##  presume linux
    sys.path.append("~/toccata/p3")
else:  ##  presume windows
    sys.path.append("c:\\toccata\\p3")
    

from globdefs import *
from glib import  add2dic, numerize
from glib import  showobj
from tocclib import  corsums, corrval
from tocclib import  bastats, midvalue
from tocclib import  ranx


##  global vars :
Showoff = 1  ##  for debugging o/p
Gold = (1.0+5**0.5)*0.5  ##  good approximation to golden number
Maxtoks = 16 * 1024  ##  maximum # of tokens 2consider from each document


##  functions first, then class(es) :

##  GETFREX :
def getfrex (toklist):
    "makes freqency table for a single token list."

    freqtab = {}
    for t in toklist:
        f = freqtab.get(t,0)
        freqtab[t] = f + 1

    return  freqtab


##  TOPTOKS :
def toptoks (freqtab,h=144):
    "returns set of high-frequency tokens from frequency dict."

    v = len(freqtab)  ##  vocsize
    sortlist = []
    for k in freqtab:
        f = freqtab[k]
        if f > 1:  ## always skip hapaxes
            sortlist.append((f,k))
    sortlist.sort(reverse=True)

    return  sortlist[:h]


##  TOKSEEK :
def tokseek (doclist,toks2get=144,multivox=1,mindocf=2):
    "returns list of tokens with high document frequencies."
    
    ttab = {}
    docs = len(doclist)  ##  may be a subsample, so recompute size
    for j,d in enumerate(doclist):
        tset = set(d.toklist[:Maxtoks])
        for t in tset:
            if t in ttab:
                ttab[t] += 1
            else:  ##  novelty
                ttab[t] = 1

    ##  sort and prune ttab :
    sortlist = []
    for t in ttab:
        tc = ttab[t]
        if tc < mindocf:
            continue  ##  too rare to bother with
        sortlist.append((tc,-len(t),t))  ##  tiebreak by inverse size

    sortlist.sort(reverse=True)
    enuf = int(toks2get*multivox+0.5)

    return  sortlist[:enuf]


##  GETRATES :
def getrates (doclist,thistok):
    "returns occurrence rates of particular token."

    ratelist = [0.0] * len(doclist)
    for j,d in enumerate(doclist):
        fc = d.toklist.count(thistok)
        ratelist[j] = fc / d.size  ##  rate not percentage at present

    return  ratelist


####  DOCADAT :
class Docadat (object):
    ##  collection of data+methods to do document classification :
    ##  assumes access to 2 Sack() objects as args, doclist & paradat.
    ##  doclist attributes expected include: classnum, freqtab, outcome, size,
    ##    text, textlen, toklist,
    ##  paradat attributes expected include:
    ##    catlist, cats, paraline, testmeta, trainmet,

    def __init__ (self,paradat,doclist):
        "initialization method for Docadat objects."

        self.classes = paradat.cats
        self.catlist = paradat.catlist
        self.id = paradat.trainmet
        if hasattr(paradat,"testmeta"):
            self.id += paradat.testmeta
        self.alldocs = len(doclist)  ##  number of training-corpus documents

        return

    def loadpars (self,paradat,sep1=' ',sep2='='):
        "loads extra docalib parameters if any, provided in paraline."

        parline = paradat.paraline.lower().strip()
        self.pars = Sack(parline)
        parbits = parline.split(sep1)  ##  item separator defaults to space
        for pb in parbits:
            bits = pb.split(sep2)  ##  internal separator (e.g. "gramsize=4")
            if len(bits) == 2:
                v = numerize(bits[1])
                if type(v) == type(9) or type(v) == type(0.5):  ##  numeric
                    setattr(self.pars,bits[0],v)
                else:  ##  nonnumeric
                    setattr(self.pars,bits[0],bits[1])
                ##  no limits imposed !
        ##  apply some defaults :
        self.pars.slug = 1.0  ##  later 2b variable ??
        if not hasattr(self.pars,"toksort"):
            self.pars.toksort = 1  ##  whether to sort o/p tokens by quayness
        if not hasattr(self.pars,"toks2get"):
            self.pars.toks2get = 144
        if self.pars.toks2get < 1:
            self.pars.toks2get = 144
        if not hasattr(self.pars,"multivox"):  ##  extra factor b4 selection
            self.pars.multivox = Gold
        if self.pars.multivox < 1:
            self.pars.multivox = Gold
        
        return  ##  doesn't crash if paradat.paraline empty
    
    def makemods (self,paradat,doclist):
        "builds classification model for each category."

        elements = []  ##  list containing info about each selected token
        ##  first pick high-frequency tokens :
        self.freqtoks = tokseek(doclist,self.pars.toks2get,self.pars.multivox)
        self.ntox = len(self.freqtoks)
        self.ndox = len(doclist)
        ##  form tables of frequencies relative to midrates, by category :
        for tn,t in enumerate(self.freqtoks):
            element = Sack(tn)
            element.item = t[-1]  ##  last item is token
            element.docfreq = t[0]  ##  first is document freq
            element.lowvals = [0] * self.classes
            element.highvals = [0] * self.classes
            ratelist = getrates(doclist,element.item)
            element.midrate = midvalue(ratelist)  ##  more useful than mean
            ##  scan thru doclist to make contingency table :
            for j,d in enumerate(doclist):
                r = ratelist[j]
                cnum = self.catlist.index(d.outcome)
                if r > element.midrate:
                    element.highvals[cnum] += 1
                else:
                    element.lowvals[cnum] += 1
            elements.append(element)

        ##  next calculate & insert bayes factors :
        self.pars.slug = math.sqrt(self.ndox) / (2*self.classes)
        for e in elements:
            e.hibayes = [] ; e.lobayes = []
            for c in range(self.classes):
                tf = e.highvals[c] + e.lowvals[c] + 2.0*self.pars.slug
                e.hibayes.append((e.highvals[c] + self.pars.slug) / tf)
                e.lobayes.append((e.lowvals[c] + self.pars.slug) / tf)
                ##  also insert quality score :
                if sum(e.highvals) >= sum(e.lowvals):
                    bs = bastats(e.hibayes)
                else:
                    bs = bastats(e.lobayes)
                e.newquay = bs.std1  ##  s.d. of conditional probs
        ##  trim undiscriminating elements if indicated :
        if self.pars.multivox > 1:
            if len(elements) < self.pars.toks2get:
                print("Warning: only",len(elements),"elements.")
            ##  print('|',len(elements))
            elements = self.picktops(elements)
            ##  print('/',len(elements))

        self.elements = elements
        if Showoff > 1:
            self.showmods(paradat,sys.stdout)
            ##  sys.exit(0)

        return

    def picktops (self,elements):
        "returns higher-scoring fraction of ruleset."

        sortlist = []
        for epos,e in enumerate(elements):
            sortlist.append((e.newquay,e.docfreq,-len(e.item),epos))
        sortlist.sort(reverse=True)
        outlist = []
        for s in sortlist[:self.pars.toks2get]:  ##  only tops
            epos = s[-1]  ##  location in elements held as last item in tuple
            outlist.append(elements[epos])

        return  outlist

    def showmods (self,paradat,fo=sys.stdout):
        "shows info on all models."

        print("classes =",self.classes,file=fo)
        print("docs =",self.alldocs,self.ndox,file=fo)
        print("vocsize =",self.ntox,file=fo)
        print(len(self.elements),file=fo)
        print('',file=fo)
        showobj(self.pars,fo)
        print('',file=fo)
        if self.pars.toksort > 0:
            outlist = []
            for j,e in enumerate(self.elements):
                outlist.append((e.newquay,e.midrate,j))
            outlist.sort(reverse=True)
            for odat in outlist:
                showobj(self.elements[odat[-1]],fo)
            return
        ##  unsorted o/p :
        for e in self.elements:
            showobj(e,fo)

        return

    def modsims (self,thisdoc,paradat):
        "returns association score of each model to this doc."

        pvec = [1.0] * self.classes  ##  equal priors
        for j,e in enumerate(self.elements):
            ff = thisdoc.toklist.count(e.item)
            fr = ff / thisdoc.size
            for c in range(self.classes):
                if fr > e.midrate:
                    pvec[c] *= e.hibayes[c]
                else:
                    pvec[c] *= e.lobayes[c]

        ##  attenuate (aways too extreme) :
        pvec = [p**0.5 for p in pvec]
        ##  rescale to 1.0 :
        psum = sum(pvec)
        return  [p/psum for p in pvec]
        ##  quasi-probs, so bigger means more likely.
