"Python3 document categorization library, topvocs method."

##  DOCALIB :
##
##  Python3 data & routines to supply to toccata test harness s/w :
##  Copyright 2013, R.S. Forsyth.
##  Released under GNU Public licence, version 3:
##  http://www.gnu.org/licenses/gpl.txt
##  first version : 16/01/2013
##  last revision : 26/06/2015


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)

import  math, os, string, sys

if sys.platform.startswith("linu"):  ##  presume linux
    sys.path.append("~/toccata/p3")
else:  ##  presume windows
    sys.path.append("c:\\toccata\\p3")
    

from globdefs import *
from glib import  add2dic, numerize
from tocclib import  corsums, corrval
from tocclib import  ranx


##  global vars :
Showoff = 0  ##  >0 for debugging o/p
Gold = 5.0 ** 0.5 * 0.5 + 0.5  ##  good approximation to golden number


##  functions first, then class(es) :

##  GETFREX :
def getfrex (toklist):
    "makes freqency table for a single token list."

    freqtab = {}
    for t in toklist:
        f = freqtab.get(t,0)
        freqtab[t] = f + 1

    return  freqtab


##  TOPTOKS :
def toptoks (freqtab,h=144):
    "returns set of high-frequency tokens from frequency dict."

    v = len(freqtab)  ##  vocsize
    sortlist = []
    for k in freqtab:
        f = freqtab[k]
        if f > 1:  ## always skip hapaxes
            sortlist.append((f,k))
    sortlist.sort(reverse=True)

    return  sortlist[:h]



####  DOCADAT :
class Docadat (object):
    ##  collection of data+methods to do document classification :
    ##  assumes access to 2 Sack() objects as args, doclist & paradat.
    ##  doclist attributes expected include:
    ##    classnum, freqtab, outcome, size, text, textlen, toklist,
    ##  paradat attributes expected include:
    ##    catlist, cats, paraline, testmeta, trainmet,

    def __init__ (self,paradat,doclist):
        "initialization method for Docadat objects."

        self.classes = paradat.cats
        self.catlist = paradat.catlist
        self.id = paradat.trainmet+paradat.testmeta
        self.alldocs = len(doclist)  ##  number of training-corpus documents

        return

    def loadpars (self,paradat,sep1=' ',sep2='='):
        "loads extra docalib parameters if any, provided in paraline."

        parline = paradat.paraline.lower().strip()
        self.pars = Sack(parline)
        parbits = parline.split(sep1)  ##  item separator defaults to space
        for pb in parbits:
            bits = pb.split(sep2)  ##  internal separator (e.g. "gramsize=4")
            if len(bits) == 2:
                v = numerize(bits[1])
                if type(v) == type(9) or type(v) == type(0.5):  ##  numeric
                    setattr(self.pars,bits[0],v)
                else:  ##  nonnumeric
                    setattr(self.pars,bits[0],bits[1])
                ##  no limits imposed !
        ##  determine correlation mode :
        self.pars.rankcor = 1  ##  default is Spearman's rank corr
        cm = "ra"
        if hasattr(self.pars,"corrmode"):
            cm = getattr(self.pars,"corrmode","ra")
        elif hasattr(self.pars,"cormode"):  ##  allow misspelling
            cm = self.pars.cormode
        if cm != "ra":  ##  must be Pearson's r
            self.pars.rankcor = 0
        
        return  ##  doesn't crash if paradat.paraline empty
    
    def makemods (self,paradat,doclist):
        "builds classification models for each category."

        self.modlist = []  ##  list of category models
        self.modeltab = {}  ##  dictionary of models, if lookup by name needed
        self.name2num = {}
        self.fullfrex = {}  ##  overall training-corpus frequency table
        cn = 0
        for c in self.catlist:  ##  relies on consistent order of catlist
            m = self.makemod(c,cn,paradat,doclist)
            self.modlist.append(m)
            self.modeltab[c] = m  ##  also point from dict keyed by catname
            self.name2num[c] = cn
            add2dic(self.fullfrex,m.freqtab)
            cn += 1  ##  keep track of position in model list
        
        self.mods = len(self.modlist)
        assert  self.mods == self.classes
        self.docs = len(doclist)
        self.vocsize = len(self.fullfrex)  ##  vocabulary size
        self.tops2use = int(self.vocsize**0.5 + 0.5)  ##  desired number of features
        self.alltops = self.tops2use  ##  ditto (12/08/2013)
        ##  alltops no longer differs from tops2use.
        self.toplist = toptoks(self.fullfrex,self.alltops)
        ##  high-frequency vocs, i.e. features, plus some spares.
        if Showoff > 0:
            print(self.alltops,self.tops2use,file=paradat.logfile)
            print(self.toplist[:self.tops2use],'\n',file=paradat.logfile)
        
        return

    def makemod (self,c,cn,paradat,doclist):
        "builds model info for a single category."

        m = Sack(c)  ##  gathered into a Sack() object
        m.freqtab = {}
        m.tottoks = 0
        m.catname = c
        m.classnum = cn
        for d in doclist:
            if d.outcome == c:
                add2dic(m.freqtab,d.freqtab)
                ##  freqtabs expected 2b computed by calling program.
                ##  safer for this library not2 cause side-effects in caller.
                m.tottoks += d.size
        m.vocsize = len(m.freqtab)
        m.tops2use = int(m.vocsize ** 0.5 + 0.5)  ##  desired number of features
        m.alltops = m.tops2use  ##  ditto (12/08/2013)
        ##  m.toplist = toptoks(m.freqtab,m.alltops)  ##  redundant ??
                                   
        return  m

    def showmod (self,cn,fo=sys.stdout):
        "shows info related to model of category number cn."

        m = self.modlist[cn]
        assert cn == m.classnum
        print("category:",m.catname,'(',cn,')',file=fo)
        print("tokens =",m.tottoks,file=fo)
        print("vocsize =",len(m.freqtab),file=fo)
        keyterms = self.toplist  ##  should have tops2use items
        for j,k in enumerate(keyterms):
            mf = m.freqtab.get(k[-1],0)
            r = mf / m.tottoks  ##  user wd want 2c this
            ##  print(j,k[-1],mf,round(r*100.0,4),file=fo)
            lout = "{0:4d} {1:20s} {2:5d} {3:8.4f}".format(j,k[-1],mf,r*100.0)
            print(lout,file=fo)
        print('',file=fo)

        return

    def showmods (self,paradat,fo=sys.stdout):
        "shows models of all categories."

        print("classes =",self.classes,file=fo)
        print(paradat.catlist,file=fo)
        print("docs =",paradat.docs,file=fo)
        print("tokens =",paradat.tottoks,file=fo)
        ##  assert paradat.tottoks == sum(self.fullfrex.values())
        print("vocsize =",self.vocsize,file=fo)
        print("topterms =",self.tops2use,file=fo)
        print("rankcor =",self.pars.rankcor,NL,file=fo)
        for cn in range(self.classes):
            self.showmod(cn,fo)

        return
        ##  doesn't open or close o/p file; caller must do that.

    def matchmod (self,cn,thisdoc,paradat):
        "matches model for category number cn to given document."

        keyterms = self.toplist[:self.tops2use]  ##  only use high chunk
        nk = len(keyterms)
        if Showoff > 0:
            print(nk,cn,thisdoc.name,file=paradat.logfile)
        xvec = [0] * nk ; yvec = [0] * nk
        for j,k in enumerate(keyterms):
            xvec[j] = self.modlist[cn].freqtab.get(k[-1],0)
            yvec[j] = thisdoc.freqtab.get(k[-1],0)
            ##  last item of tuple k is expected 2b token itself.
            if Showoff > 0 and j < 26:
                print(j,k,xvec[j],yvec[j],file=paradat.logfile)  ##  checking
        if self.pars.rankcor > 0:
            cx = corsums(ranx(xvec),ranx(yvec))
        else:
            cx = corsums(xvec,yvec)
        
        return  corrval(cx)  ##  matching score, higher implies stronger match

    def modsims (self,thisdoc,paradat):
        "returns association score of each model to this doc."

        catvals = [0.0] * paradat.cats
        for cpos,c in enumerate(paradat.catlist):
            catvals[cpos] = self.matchmod(cpos,thisdoc,paradat)

        return  catvals  ##  no scaling to quasi-probs at present.
        ##  also no inversion if distances used, so more means better.
